// quickSort con comparatore
def quickSort[T](cmp:(T,T)=>Int)(l:List[T]):List[T] = {
	if (l.isEmpty) l
	else {
		val sx = l.filter(x=>cmp(x,l.head)<0)
		val dx = l.filter(x=>cmp(x,l.head)>=0).tail
		quickSort(cmp)(sx):::(l.head::quickSort(cmp)(dx))
	}
}

// scrivere una variante della funzione flatten
// che chiameremo myFlatten definita usando i metodi
// sulle liste
def myFlatten[T](l:List[List[T]]):List[T] = {
	l.reduce(_:::_)
}

// algoritmo di ordinamento per inserimento
// (insertionSort). Ci serve come ingrediente
// un metodo che, dato un elemento x e una lista l
// ordinata, inserisce x nella posizione che gli
// compete in l.
def insert(x:Int,l:List[Int]):List[Int] = {
	if (l.isEmpty) List(x)
	else if (x<l.head) x::l
	else l.head::insert(x,l.tail)
}

def insertionSort(l:List[Int]):List[Int] = {
	if (l.size<2) l
	else insert(l.head,insertionSort(l.tail))
}

// calcolo del prodotto cartesiano {1,n}x{1,n}
// prod(n) = List((1,1), (1,2), ... (1,n),
//                (2,1), (2,2), ... (2,n),
//                ...
//                (n,1), (n,2), ... (n,n))
def prod(n:Int) = {
	(1 to n).map(x=>
		(1 to n).map(y=>(x,y))).toList.flatten
}

// variante usando il costrutto for comprehension
// che è un modo per generare collezioni di elementi
def prod2(n:Int) = {
	val l = for {
		x <- 1 to n
		y <- 1 to n
	} yield (x,y)
	l.toList
}

// variante che genera triple
def prod3(n:Int) = {
	val l = for {
		x <- 1 to n
		y <- 1 to n
		z <- 1 to n
	} yield (x,y,z)
	l.toList
}

// variante di prod2 con passo 2 sugli indici
def prod2_passo2(n:Int) = {
	val l = for {
		x <- 1 to n
		y <- 1 to n
	} yield (2*x, 2*y)
	l.toList
}

// variante di prod3 usando map e flatten
def prod3_variante(n:Int) = {
	(1 to n).map(x=>
		(1 to n).map(y=>
			(1 to n).map(z=>(x,y,z)))).
				flatten.flatten.toList
}

// indici coppie matrice triangolare
def matriceTriangolare(n:Int) = {
	val l = for {
		i <- 1 to n
		j <- i to n
	} yield (i,j)
	l.toList
}

// variante
def matriceTriangolare2(n:Int) = {
	val l = for {
		i <- 1 to n
		j <- 1 to n
		if (j>=i)
	} yield (i,j)
	l.toList
}

// scrivere un metodo che, data una lista l di elementi
// di tipo T e un predicato che dato un elemento di
// tipo T restituisce vero o falso, filtra tutti gli
// elementi di l che soddisfano il predicato
// Il metodo deve essere ricorsivo
def myFilter[T](l:List[T],p:T=>Boolean):List[T] = {
	if (l.isEmpty) l
	else if (p(l.head)) l.head :: myFilter(l.tail,p)
	else myFilter(l.tail,p)
}

// scrivere una variante ricorsiva del metodo
// forall delle liste
def myForall[T](l:List[T], p:T=>Boolean):Boolean = {
	if (l.isEmpty) true
	else p(l.head) && myForall(l.tail,p)
}

// scrivere una variante ricorsiva del metodo
// exists delle liste
def myExists[T](l:List[T], p:T=>Boolean):Boolean = {
	if (l.isEmpty) false
	else p(l.head) || myExists(l.tail,p)
}

// pattern matching su costanti
def f(n:Int) = {
	n match {
		case 0 => "zero"
		case 1 => "uno"
		case 2 => "due"
		case _ => "non so"
	}
}

// altro esempio, diverso da quello che sarebbe
// realizzabile con un costrutto come lo switch
// di C o Java
def f(s:String) = {
	s.toLowerCase match {
		case "zero" => 0
		case "uno" => 1
		case "due" => 2
	}
}

// match multiplo su costanti
def verificaPariDispari(n:Int) = {
	n % 10 match {
		case 0|2|4|6|8 => ""+n+" è pari"
		case _ => ""+n+" è dispari"
	}
}

// matching su variabili
def f(x:Any) = {
	x match {
		case i:Int => ""+i+" è Int"
		case f:Float => ""+f+" è Float"
		case d:Double => ""+d+" è Double"
		case s:String => ""+s+" è String"
		case _ => "non so"
	}
}

// se non ci interessa il valore della variabile n
// ma solo il suo tipo, si può scrivere
def f(x:Any) = x match {
	case _:Int => "un intero"
	case _:Float => "un float"
	case _:String => "una String"
	case _ => "tipo ignoto"
}

// matching con condizioni
def f(n:Int) = n match {
	case x if (x<0) => "negativo"
	case _ => "non negativo"
}

// oppure
def f(n:Int) = n match {
	case _ if (n<0) => "negativo"
	case _ => "non negativo"
}

// matching su liste
// Nota: h vale l.head e t vale l.tail
// Nota 2: se concateno h con t ottengo l che è la
//         lista su cui sto facendo in match
def mySize[T](l:List[T]):Int = l match {
	case Nil => 0
	case h::t => 1 + mySize(t)
}

// matching su coppie
// riprendiamo l'esempio della merge di due liste
// ordinate
def merge2(l1:List[Int], l2:List[Int]):List[Int] = {
	(l1,l2) match {
		case (Nil,Nil) => Nil
		case (l,Nil) => l
		case (Nil,l) => l
		case (h1::t1,h2::t2) if (h1<h2) =>
			 					h1::merge2(t1,l2)
		case (h1::t1,h2::t2) => h2::merge2(l1,t2)
	}
}

// altro esempio sulle liste
// verifica se una lista è ordinata
def ordinata(l:List[Int]):Boolean = l match {
	case Nil => true
	case _::Nil => true
	case h::t if (h<=t.head) => ordinata(t)
	case _ => false
}

// altro esempio sulle liste
// versione di myMap con matching
def myMap[A,B](l:List[A],f:A=>B):List[B] = l match {
	case Nil => Nil
	case h::t => f(h)::myMap(t,f)
}
